from ph_adorb.fuel import PhAdorbFuel, PhAdorbFuelType


def test_fuel():
    fuel = PhAdorbFuel(
        fuel_type=PhAdorbFuelType.ELECTRICITY,
        purchase_price_per_kwh=0.12,
        sale_price_per_kwh=0.08,
        annual_base_price=0.10,
        used=True,
    )
    assert fuel.name == "Electricity"
    assert fuel.purchase_price_per_kwh == 0.12
    assert fuel.sale_price_per_kwh == 0.08
    assert fuel.annual_base_price == 0.10


def test_fuel_to_json():
    fuel = PhAdorbFuel(
        fuel_type=PhAdorbFuelType.ELECTRICITY,
        purchase_price_per_kwh=0.12,
        sale_price_per_kwh=0.08,
        annual_base_price=0.10,
        used=True,
    )
    json_str = fuel.model_dump_json()
    assert (
        json_str
        == '{"fuel_type":"Electricity","purchase_price_per_kwh":0.12,"sale_price_per_kwh":0.08,"annual_base_price":0.1,"used":true}'
    )


def test_fuel_from_json():
    fuel = PhAdorbFuel.model_validate(
        {
            "fuel_type": "Electricity",
            "purchase_price_per_kwh": 0.12,
            "sale_price_per_kwh": 0.08,
            "annual_base_price": 0.10,
            "used": True,
        }
    )
    assert fuel.fuel_type == PhAdorbFuelType.ELECTRICITY
    assert fuel.purchase_price_per_kwh == 0.12
    assert fuel.sale_price_per_kwh == 0.08
    assert fuel.annual_base_price == 0.10
    assert fuel.used == True
    assert fuel.name == "Electricity"



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""CO2 Reduction Measures and Collection"""

import json
from enum import Enum
from pathlib import Path

from pydantic import BaseModel, PrivateAttr


class CO2MeasureType(str, Enum):
    PERFORMANCE = "PERFORMANCE"
    NON_PERFORMANCE = "NON_PERFORMANCE"


class PhAdorbCO2ReductionMeasure(BaseModel):
    """A CO2 Reduction Measure."""

    measure_type: CO2MeasureType
    name: str
    year: int
    cost: float
    kg_CO2: float | None
    country_name: str
    labor_fraction: float

    @property
    def material_fraction(self) -> float:
        return 1.0 - self.labor_fraction


class PhAdorbCO2MeasureCollection(BaseModel):
    """A collection of CO2 Reduction Measures."""

    _measures: dict[str, PhAdorbCO2ReductionMeasure] = PrivateAttr(default_factory=dict)

    def add_measure(self, factor: PhAdorbCO2ReductionMeasure) -> None:
        self._measures[factor.name] = factor

    def get_measure(self, key: str) -> PhAdorbCO2ReductionMeasure:
        return self._measures[key]

    def keys(self) -> list[str]:
        return [k for k, v in sorted(self._measures.items(), key=lambda x: x[1].year)]

    def values(self) -> list[PhAdorbCO2ReductionMeasure]:
        return list(sorted(self._measures.values(), key=lambda x: x.year))

    def __iter__(self):
        return iter(sorted(self._measures.values(), key=lambda x: x.year))

    def __contains__(self, key: str | PhAdorbCO2ReductionMeasure) -> bool:
        if isinstance(key, PhAdorbCO2ReductionMeasure):
            return key in self._measures.values()
        return key in self._measures

    def __len__(self) -> int:
        return len(self._measures)

    @property
    def performance_measures(self) -> "PhAdorbCO2MeasureCollection":
        """Return a collection of PERFORMANCE measures."""
        new_collection = PhAdorbCO2MeasureCollection()
        for measure in self.values():
            if measure.measure_type == CO2MeasureType.PERFORMANCE:
                new_collection.add_measure(measure)
        return new_collection

    @property
    def nonperformance_measures(self) -> "PhAdorbCO2MeasureCollection":
        """Return a collection of NON-PERFORMANCE measures."""
        new_collection = PhAdorbCO2MeasureCollection()
        for measure in self.values():
            if measure.measure_type == CO2MeasureType.NON_PERFORMANCE:
                new_collection.add_measure(measure)
        return new_collection


def write_CO2_measures_to_json_file(
    _file_path: Path, measures: dict[str, PhAdorbCO2ReductionMeasure]
) -> None:
    """Write all of the CO2 Measure-Types to a JSON file."""
    with open(_file_path, "w") as json_file:
        json.dump([_.model_dump() for _ in measures.values()], json_file, indent=4)


def load_CO2_measures_from_json_file(
    _file_path: Path,
) -> dict[str, PhAdorbCO2ReductionMeasure]:
    """Load all of the CO2 Measure-Types from a JSON file."""
    with open(_file_path, "r") as json_file:
        all_measures = (
            PhAdorbCO2ReductionMeasure(**item) for item in json.load(json_file)
        )
        return {_.name: _ for _ in all_measures}



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Functions for importing Honeybee Models from HBJSON files."""

import json
import logging
import os
import pathlib
from typing import Dict

# -- Dev Note: Required to import ALL the base packages to run the __init__ startup routines
# -- which ensures that .revive properties slot is added to all HB Objects. This must be done before
# -- running read_hb_json to ensure there is a place for all the .ph properties to go.
# -----------------------------------------------------------------------------
# -- Dev Note: Do NOT remove vvvvvv -------------------------------------------
import honeybee
import honeybee_energy
import honeybee_energy_revive
import honeybee_revive
from honeybee import model

# -- Dev Note: Do NOT remove ^^^^^^ -------------------------------------------
# -----------------------------------------------------------------------------


logger = logging.getLogger()


class HBJSONModelReadError(Exception):
    def __init__(self, _in) -> None:
        self.message = (
            f"HBJSONModelReadError: Can only convert a Honeybee 'Model' to WUFI XML.\n"
            f"Got a Honeybee object of type: {_in}."
        )

        super(HBJSONModelReadError, self).__init__(self.message)


def read_hb_json_from_file(_file_address: pathlib.Path) -> Dict:
    """Read in the HBJSON file and return it as a python dictionary.

    Arguments:
    ----------
        _file_address (pathlib.Path): A valid file path for the HBJSON file to read.

    Returns:
    --------
        Dict: The HBJSON dictionary, read in from the HBJSON file.
    """

    if not os.path.isfile(_file_address):
        msg = f"FileNotFoundError: {_file_address} is not a valid file path?"
        e = FileNotFoundError(msg)
        logger.critical(e)
        raise e

    with open(_file_address) as json_file:
        data = json.load(json_file)

    if data.get("type", None) != "Model":
        e = HBJSONModelReadError(data.get("type", None))
        logger.critical(e.message)
        raise e
    else:
        return data


def convert_hbjson_dict_to_hb_model(_data: Dict) -> model.Model:
    """Convert an HBJSON python dictionary into an HB-Model

    Arguments:
    ----------
        _data (Dict): An HBJSON dictionary with all the model information.

    Returns:
    --------
        model.Model: A Honeybee Model, rebuilt from the HBJSON file.
    """
    hb_model: model.Model = model.Model.from_dict(_data)
    logger.info(f"Converting HB-Model from {hb_model.units} to Meters.")
    hb_model.convert_to_units("Meters")
    return hb_model



from pathlib import Path

from ph_adorb.measures import (
    CO2MeasureType,
    PhAdorbCO2MeasureCollection,
    PhAdorbCO2ReductionMeasure,
    load_CO2_measures_from_json_file,
    write_CO2_measures_to_json_file,
)


def test_basic_CO2ReductionMeasure():
    measure = PhAdorbCO2ReductionMeasure(
        measure_type=CO2MeasureType.PERFORMANCE,
        name="Test Measure",
        year=2023,
        cost=1000,
        kg_CO2=100,
        country_name="DE",
        labor_fraction=0.4,
    )
    assert measure.name == "Test Measure"
    assert measure.year == 2023
    assert measure.cost == 1000
    assert measure.kg_CO2 == 100
    assert measure.country_name == "DE"
    assert measure.labor_fraction == 0.4
    assert measure.material_fraction == 0.6


def test_CO2ReductionMeasure_to_json():
    measure = PhAdorbCO2ReductionMeasure(
        measure_type=CO2MeasureType.PERFORMANCE,
        name="Test Measure",
        year=2023,
        cost=1000,
        kg_CO2=100,
        country_name="DE",
        labor_fraction=0.5,
    )
    json_str = measure.model_dump_json()
    assert (
        json_str
        == '{"measure_type":"PERFORMANCE","name":"Test Measure","year":2023,"cost":1000.0,"kg_CO2":100.0,"country_name":"DE","labor_fraction":0.5}'
    )


def test_CO2ReductionMeasure_from_json():
    measure = PhAdorbCO2ReductionMeasure.model_validate(
        {
            "measure_type": "PERFORMANCE",
            "name": "Test Measure",
            "year": 2023,
            "cost": 1000,
            "kg_CO2": 100,
            "country_name": "DE",
            "labor_fraction": 0.5,
        }
    )
    assert measure.measure_type == CO2MeasureType.PERFORMANCE
    assert measure.name == "Test Measure"
    assert measure.year == 2023
    assert measure.cost == 1000
    assert measure.kg_CO2 == 100
    assert measure.country_name == "DE"
    assert measure.labor_fraction == 0.5
    assert measure.material_fraction == 0.5


def test_CO2ReductionMeasureCollection():
    collection = PhAdorbCO2MeasureCollection()
    measure1 = PhAdorbCO2ReductionMeasure(
        measure_type=CO2MeasureType.PERFORMANCE,
        name="Test Measure 1",
        year=2023,
        cost=1000,
        kg_CO2=100,
        country_name="DE",
        labor_fraction=0.5,
    )
    measure2 = PhAdorbCO2ReductionMeasure(
        measure_type=CO2MeasureType.PERFORMANCE,
        name="Test Measure 2",
        year=2024,
        cost=2000,
        kg_CO2=200,
        country_name="DE",
        labor_fraction=0.6,
    )
    collection.add_measure(measure1)
    collection.add_measure(measure2)

    assert len(collection) == 2

    assert collection.get_measure("Test Measure 1") == measure1
    assert collection.get_measure("Test Measure 2") == measure2

    assert measure1 in collection
    assert measure1 in collection.values()
    assert measure1.name in collection
    assert measure1.name in collection.keys()

    assert measure2 in collection
    assert measure2 in collection.values()
    assert measure2.name in collection
    assert measure2.name in collection.keys()


def test_CO2ReductionMeasureCollection_PERFORMANCE():
    collection = PhAdorbCO2MeasureCollection()
    measure1 = PhAdorbCO2ReductionMeasure(
        measure_type=CO2MeasureType.PERFORMANCE,
        name="Test Measure 1",
        year=2023,
        cost=1000,
        kg_CO2=100,
        country_name="DE",
        labor_fraction=0.5,
    )
    measure2 = PhAdorbCO2ReductionMeasure(
        measure_type=CO2MeasureType.PERFORMANCE,
        name="Test Measure 2",
        year=2024,
        cost=2000,
        kg_CO2=200,
        country_name="DE",
        labor_fraction=0.6,
    )
    measure3 = PhAdorbCO2ReductionMeasure(
        measure_type=CO2MeasureType.NON_PERFORMANCE,
        name="Test Measure 3",
        year=2024,
        cost=2000,
        kg_CO2=200,
        country_name="DE",
        labor_fraction=0.6,
    )
    collection.add_measure(measure1)
    collection.add_measure(measure2)
    collection.add_measure(measure3)

    assert len(collection) == 3
    assert len(collection.performance_measures) == 2
    assert len(collection.nonperformance_measures) == 1


def test_CO2ReductionMeasures_json_file():
    collection = PhAdorbCO2MeasureCollection()
    measure1 = PhAdorbCO2ReductionMeasure(
        measure_type=CO2MeasureType.PERFORMANCE,
        name="Test Measure 1",
        year=2023,
        cost=1000,
        kg_CO2=100,
        country_name="DE",
        labor_fraction=0.5,
    )
    measure2 = PhAdorbCO2ReductionMeasure(
        measure_type=CO2MeasureType.PERFORMANCE,
        name="Test Measure 2",
        year=2024,
        cost=2000,
        kg_CO2=200,
        country_name="DE",
        labor_fraction=0.6,
    )
    collection.add_measure(measure1)
    collection.add_measure(measure2)

    file_path = Path("test_measures.json")
    write_CO2_measures_to_json_file(file_path, {_.name: _ for _ in collection})

    # -- Read the JSON file data back in
    measures = load_CO2_measures_from_json_file(file_path)
    assert len(measures) == 2
    assert measures["Test Measure 1"] == measure1
    assert measures["Test Measure 2"] == measure2

    # -- Clean-up
    file_path.unlink()



from pathlib import Path

from ph_adorb.equipment import (
    PhAdorbEquipment,
    PhAdorbEquipmentCollection,
    PhAdorbEquipmentType,
    load_equipment_from_json_file,
    write_equipment_to_json_file,
)


def test_basic_equipment():
    equip = PhAdorbEquipment(
        name="Test Equipment",
        equipment_type=PhAdorbEquipmentType.MECHANICAL,
        cost=1000.0,
        lifetime_years=10,
        labor_fraction=0.2,
    )
    assert equip.name == "Test Equipment"
    assert equip.equipment_type == PhAdorbEquipmentType.MECHANICAL
    assert equip.cost == 1000.0
    assert equip.lifetime_years == 10
    assert equip.labor_fraction == 0.2
    assert equip.material_fraction == 0.8


def test_equipment_duplicate():
    equip = PhAdorbEquipment(
        name="Test Equipment",
        equipment_type=PhAdorbEquipmentType.MECHANICAL,
        cost=1000.0,
        lifetime_years=10,
        labor_fraction=0.2,
    )
    equip2 = equip.duplicate()
    assert equip is not equip2
    assert equip == equip2


def test_equipment_copy():
    equip = PhAdorbEquipment(
        name="Test Equipment",
        equipment_type=PhAdorbEquipmentType.MECHANICAL,
        cost=1000.0,
        lifetime_years=10,
        labor_fraction=0.2,
    )
    equip2 = equip.__copy__()
    assert equip is not equip2
    assert equip == equip2


def test_equipment_to_json():
    equip = PhAdorbEquipment(
        name="Test Equipment",
        equipment_type=PhAdorbEquipmentType.MECHANICAL,
        cost=1000.0,
        lifetime_years=10,
        labor_fraction=0.2,
    )
    json_str = equip.model_dump_json()
    assert (
        json_str
        == '{"name":"Test Equipment","equipment_type":"Mechanical","cost":1000.0,"lifetime_years":10,"labor_fraction":0.2}'
    )


def test_equipment_from_json():
    json_str = '{"name": "Test Equipment", "equipment_type": "Mechanical", "cost": 1000.0, "lifetime_years": 10, "labor_fraction": 0.2}'
    equip = PhAdorbEquipment.model_validate_json(json_str)
    assert equip.name == "Test Equipment"
    assert equip.equipment_type == PhAdorbEquipmentType.MECHANICAL
    assert equip.cost == 1000.0
    assert equip.lifetime_years == 10
    assert equip.labor_fraction == 0.2
    assert equip.material_fraction == 0.8


def test_EquipmentCollection():
    equip1 = PhAdorbEquipment(
        name="Test Mechanical",
        equipment_type=PhAdorbEquipmentType.MECHANICAL,
        cost=1000.0,
        lifetime_years=10,
        labor_fraction=0.2,
    )
    equip2 = PhAdorbEquipment(
        name="Test Lights",
        equipment_type=PhAdorbEquipmentType.LIGHTS,
        cost=2000.0,
        lifetime_years=20,
        labor_fraction=0.3,
    )
    equip3 = PhAdorbEquipment(
        name="Test Appliance",
        equipment_type=PhAdorbEquipmentType.APPLIANCE,
        cost=3000.0,
        lifetime_years=30,
        labor_fraction=0.4,
    )
    equip_collection = PhAdorbEquipmentCollection()
    equip_collection.add_equipment(equip1)
    equip_collection.add_equipment(equip2)
    equip_collection.add_equipment(equip3)

    assert len(equip_collection) == 3
    assert equip_collection.get_equipment(equip1.name) == equip1
    assert equip_collection.get_equipment(equip2.name) == equip2
    assert equip_collection.get_equipment(equip3.name) == equip3

    # -- Test Keys and Values, Contains
    assert equip1 in equip_collection
    assert equip1 in equip_collection.values()
    assert equip1.name in equip_collection
    assert equip1.name in equip_collection.keys()

    assert equip2 in equip_collection
    assert equip2 in equip_collection.values()
    assert equip2.name in equip_collection
    assert equip2.name in equip_collection.keys()

    assert equip3 in equip_collection
    assert equip3 in equip_collection.values()
    assert equip3.name in equip_collection
    assert equip3.name in equip_collection.keys()

    # -- Test Iteration
    for equip in equip_collection:
        assert isinstance(equip, PhAdorbEquipment)


def test_equipment_json_file():
    equip1 = PhAdorbEquipment(
        name="Test Mechanical",
        equipment_type=PhAdorbEquipmentType.MECHANICAL,
        cost=1000.0,
        lifetime_years=10,
        labor_fraction=0.2,
    )
    equip2 = PhAdorbEquipment(
        name="Test Lights",
        equipment_type=PhAdorbEquipmentType.LIGHTS,
        cost=2000.0,
        lifetime_years=20,
        labor_fraction=0.3,
    )
    equip3 = PhAdorbEquipment(
        name="Test Appliance",
        equipment_type=PhAdorbEquipmentType.APPLIANCE,
        cost=3000.0,
        lifetime_years=30,
        labor_fraction=0.4,
    )

    file_path = Path("test_equipment.json")
    write_equipment_to_json_file(
        file_path, {_.name: _ for _ in [equip1, equip2, equip3]}
    )

    # -- Read the JSON file back in
    equipment = load_equipment_from_json_file(file_path)
    assert len(equipment) == 3
    assert equipment["Test Mechanical"] == equip1
    assert equipment["Test Lights"] == equip2
    assert equipment["Test Appliance"] == equip3

    # -- Cleanup
    file_path.unlink()



from copy import copy
from pathlib import Path

from pytest import approx

from ph_adorb.constructions import (
    PhAdorbConstruction,
    PhAdorbConstructionCollection,
    load_constructions_from_json_file,
    write_constructions_to_json_file,
)


def test_basic_construction():
    construction = PhAdorbConstruction(
        display_name="Test Construction",
        identifier="Test Construction",
        CO2_kg_per_m2=100,
        cost_per_m2=1000,
        lifetime_years=30,
        labor_fraction=0.4,
    )
    assert construction.display_name == "Test Construction"
    assert construction.CO2_kg_per_m2 == 100
    assert construction.cost_per_m2 == 1000
    assert construction.lifetime_years == 30
    assert construction.labor_fraction == 0.4
    assert construction.material_fraction == 0.6


def test_set_quantity_ft2():
    construction = PhAdorbConstruction(
        display_name="Test Construction",
        identifier="Test Construction",
        CO2_kg_per_m2=100,
        cost_per_m2=1000,
        lifetime_years=30,
        labor_fraction=0.4,
    )
    construction.set_quantity_ft2(1076.39)
    assert construction.area_m2 == approx(100.0)
    assert construction.quantity_ft2 == approx(1076.39)


def test_cost():
    construction = PhAdorbConstruction(
        display_name="Test Construction",
        identifier="Test Construction",
        CO2_kg_per_m2=100,
        cost_per_m2=1000,
        lifetime_years=30,
        labor_fraction=0.4,
    )
    construction.area_m2 = 100
    assert construction.cost == approx(100000.0)


def test_CO2_kg():
    construction = PhAdorbConstruction(
        display_name="Test Construction",
        identifier="Test Construction",
        CO2_kg_per_m2=100,
        cost_per_m2=1000,
        lifetime_years=30,
        labor_fraction=0.4,
    )
    construction.area_m2 = 100
    assert construction.CO2_kg == approx(10000.0)


def test_construction_to_json():
    construction = PhAdorbConstruction(
        display_name="Test Construction",
        identifier="Test Construction",
        CO2_kg_per_m2=100,
        cost_per_m2=1000,
        lifetime_years=30,
        labor_fraction=0.4,
    )
    json_str = construction.model_dump_json()
    assert (
        json_str
        == '{"display_name":"Test Construction","identifier":"Test Construction","CO2_kg_per_m2":100.0,"cost_per_m2":1000.0,"lifetime_years":30,"labor_fraction":0.4,"area_m2":0.0}'
    )


def test_construction_from_json():
    construction = PhAdorbConstruction.model_validate(
        {
            "display_name": "Test Construction",
            "identifier": "Test Construction",
            "CO2_kg_per_m2": 100,
            "cost_per_m2": 1000,
            "lifetime_years": 30,
            "labor_fraction": 0.4,
        }
    )
    assert construction.display_name == "Test Construction"
    assert construction.identifier == "Test Construction"
    assert construction.CO2_kg_per_m2 == 100
    assert construction.cost_per_m2 == 1000
    assert construction.lifetime_years == 30
    assert construction.labor_fraction == 0.4
    assert construction.material_fraction == 0.6


def test_duplicate_construction():
    construction = PhAdorbConstruction(
        display_name="Test Construction",
        identifier="Test Construction",
        CO2_kg_per_m2=100,
        cost_per_m2=1000,
        lifetime_years=30,
        labor_fraction=0.4,
    )

    construction2 = construction.duplicate()
    assert construction is not construction2
    assert construction == construction2

    construction3 = copy(construction)
    assert construction is not construction3
    assert construction == construction3


def test_construction_collection():
    collection = PhAdorbConstructionCollection()
    construction1 = PhAdorbConstruction(
        display_name="Test Construction 1",
        identifier="Test Construction 1",
        CO2_kg_per_m2=100,
        cost_per_m2=1000,
        lifetime_years=30,
        labor_fraction=0.4,
    )
    construction2 = PhAdorbConstruction(
        display_name="Test Construction 2",
        identifier="Test Construction 1",
        CO2_kg_per_m2=200,
        cost_per_m2=2000,
        lifetime_years=40,
        labor_fraction=0.5,
    )
    collection.add_construction(construction1)
    collection.add_construction(construction2)

    assert len(collection) == 2

    # Test Getting Construction
    assert collection.get_construction("Test Construction 1") == construction1
    assert collection.get_construction("Test Construction 2") == construction2

    # Test Keys, Values, Contains
    assert "Test Construction 1" in collection
    assert "Test Construction 1" in collection.keys()
    assert construction1 in collection
    assert construction1 in collection.values()

    assert "Test Construction 2" in collection
    assert "Test Construction 2" in collection.keys()
    assert construction2 in collection
    assert construction2 in collection.values()

    # Test iteration
    for construction in collection:
        assert isinstance(construction, PhAdorbConstruction)


def test_set_construction_f2_quantities():
    collection = PhAdorbConstructionCollection()
    construction1 = PhAdorbConstruction(
        display_name="Test Construction 1",
        identifier="Test Construction 1",
        CO2_kg_per_m2=100,
        cost_per_m2=1000,
        lifetime_years=30,
        labor_fraction=0.4,
    )
    construction2 = PhAdorbConstruction(
        display_name="Test Construction 2",
        identifier="Test Construction 1",
        CO2_kg_per_m2=200,
        cost_per_m2=2000,
        lifetime_years=40,
        labor_fraction=0.5,
    )
    collection.add_construction(construction1)
    collection.add_construction(construction2)

    quantities = {"TEST CONSTRUCTION 1": 1076.39, "TesT construction 2": 1076.39}
    collection.set_constructions_ft2_quantities(quantities)
    assert collection.get_construction("Test Construction 1").area_m2 == approx(100.0)
    assert collection.get_construction("Test Construction 2").area_m2 == approx(100.0)


def test_constructions_json_file():
    # -- Create a temp JSON file with some constructions
    c1 = PhAdorbConstruction(
        display_name="Test Construction 1",
        identifier="Test Construction 1",
        CO2_kg_per_m2=100,
        cost_per_m2=1000,
        lifetime_years=30,
        labor_fraction=0.4,
    )
    c2 = PhAdorbConstruction(
        display_name="Test Construction 2",
        identifier="Test Construction 1",
        CO2_kg_per_m2=200,
        cost_per_m2=2000,
        lifetime_years=40,
        labor_fraction=0.5,
    )
    constructions = [c1, c2]

    file_path = Path("temp.json")
    write_constructions_to_json_file(
        file_path, {_.display_name: _ for _ in constructions}
    )

    # -- Read the JSON file back in
    constructions2 = load_constructions_from_json_file(file_path)
    assert len(constructions2) == 2
    assert constructions2["Test Construction 1"] == c1
    assert constructions2["Test Construction 2"] == c2

    # -- Clean up
    file_path.unlink()



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Functions to load and process the source data SQL files."""

import sqlite3
from collections import defaultdict
from pathlib import Path

from ph_units.unit_type import Unit
from pydantic import BaseModel

KWH_PER_JOULE = 0.0000002778


class DataFileSQL(BaseModel):
    """A single EnergyPlus results .SQL Data File."""

    source_file_path: Path

    @property
    def file_name(self) -> str:
        """The name of the file."""
        return self.source_file_path.name

    def get_peak_electric_watts(self) -> float:
        """Get the 'Facility Total Building Electricity Demand Rate' [W] from the SQL File."""
        conn = sqlite3.connect(self.source_file_path)
        try:
            c = conn.cursor()
            # Note: I am not sure which of these two is the right one to use.
            # c.execute(
            #     "SELECT MAX(Value) FROM 'ReportVariableWithTime' "
            #     "WHERE Name='Facility Total Building Electricity Demand Rate'"
            # )
            c.execute(
                "SELECT Value FROM TabularDataWithStrings WHERE ReportName='DemandEndUseComponentsSummary' "
                "AND ColumnName='Electricity' AND RowName='Total End Uses'"
            )
            peak_electric_watts_ = c.fetchone()[0]
        except Exception as e:
            conn.close()
            raise Exception(str(e))
        finally:
            conn.close()

        return peak_electric_watts_

    def get_hourly_purchased_electricity_kwh(self) -> list[float]:
        """Get the 'Facility Total Building Electricity Demand Rate' [W] from the SQL File."""
        conn = sqlite3.connect(self.source_file_path)
        try:
            c = conn.cursor()
            c.execute(
                "SELECT Value FROM 'ReportVariableWithTime' " "WHERE Name='Facility Total Purchased Electricity Energy'"
            )
            total_purchased_electricity_kwh_ = [_[0] * KWH_PER_JOULE for _ in c.fetchall()]
        except Exception as e:
            conn.close()
            raise Exception(str(e))
        finally:
            conn.close()

        return total_purchased_electricity_kwh_

    def get_total_purchased_electricity_kwh(self) -> float:
        """Get the 'Facility Total Building Electricity Demand Rate' [W] from the SQL File."""
        conn = sqlite3.connect(self.source_file_path)
        try:
            c = conn.cursor()
            c.execute(
                "SELECT SUM(Value) FROM 'ReportVariableWithTime' "
                "WHERE Name='Facility Total Purchased Electricity Energy'"
            )
            total_purchased_electricity_kwh_ = c.fetchone()[0] * KWH_PER_JOULE
        except Exception as e:
            conn.close()
            raise Exception(str(e))
        finally:
            conn.close()

        return total_purchased_electricity_kwh_

    def get_total_sold_electricity_kwh(self) -> float:
        """Get the 'Facility Total Building Electricity Demand Rate' [W] from the SQL File."""
        conn = sqlite3.connect(self.source_file_path)
        try:
            c = conn.cursor()
            c.execute(
                "SELECT SUM(Value) FROM 'ReportVariableWithTime' "
                "WHERE Name='Facility Total Surplus Electricity Energy'"
            )
            total_sold_electricity_kwh_ = c.fetchone()[0] * KWH_PER_JOULE
        except Exception as e:
            conn.close()
            raise Exception(str(e))
        finally:
            conn.close()

        return total_sold_electricity_kwh_

    def get_total_purchased_gas_kwh(self) -> float:
        """Return the total purchased gas in KWH."""
        fuel_use_dict = self.get_total_end_kwh_by_fuel_type()
        return fuel_use_dict["Natural Gas"]

    def get_total_end_kwh_by_fuel_type(self) -> dict[str, float]:
        # -- Get the data from the SQL file
        conn = sqlite3.connect(self.source_file_path)
        try:
            c = conn.cursor()
            c.execute(
                "SELECT ColumnName, RowName, Value, Units FROM TabularDataWithStrings "
                "WHERE TableName='End Uses By Subcategory' AND ReportName='AnnualBuildingUtilityPerformanceSummary'"
            )
            table_data = c.fetchall()
            conn.close()
        except Exception as e:
            conn.close()
            raise Exception(str(e))
        finally:
            conn.close()

        # -- Collect all the SQL data into a dictionary
        table_dict = defaultdict(dict)
        for item in table_data:
            fuel_type_name, end_use_name, value, unit = item
            try:
                table_dict[fuel_type_name][end_use_name] = Unit(value, unit).as_a("KWH").value
            except ValueError:
                # -- Not an energy number (m3/s, etc...)
                pass

        # -- Total the data by fuel-type
        energy_by_fuel_type = {}
        for fuel_type, end_use_data in table_dict.items():
            energy_by_fuel_type[fuel_type] = sum(end_use_data.values())

        return energy_by_fuel_type



from pathlib import Path

from ph_adorb.grid_region import (
    PhAdorbGridRegion,
    load_CO2_factors_from_json_file,
    write_CO2_factors_to_json_file,
)


def test_GridRegionFactors():
    hourly_CO2_factors = {
        2023: [460.1, 469.3],
        2024: [460.1, 475.3],
        2025: [434.1, 445.2],
    }
    grid_region_factors = PhAdorbGridRegion(
        region_code="DE",
        region_name="Germany",
        description="Germany",
        hourly_CO2_factors=hourly_CO2_factors,
    )
    assert grid_region_factors.region_code == "DE"
    assert grid_region_factors.hourly_CO2_factors == hourly_CO2_factors
    assert grid_region_factors.get_CO2_factors_as_df().shape == (2, 3)


def test_GridRegionFactors_to_json():
    hourly_CO2_factors = {
        2023: [460.1, 469.3],
        2024: [460.1, 475.3],
        2025: [434.1, 445.2],
    }
    grid_region_factors = PhAdorbGridRegion(
        region_code="DE",
        region_name="Germany",
        description="Germany",
        hourly_CO2_factors=hourly_CO2_factors,
    )
    json_str = grid_region_factors.model_dump_json()
    assert (
        json_str
        == '{"region_code":"DE","region_name":"Germany","description":"Germany","hourly_CO2_factors":{"2023":[460.1,469.3],"2024":[460.1,475.3],"2025":[434.1,445.2]}}'
    )


def test_GridRegionFactors_from_json():
    hourly_CO2_factors = {
        2023: [460.1, 469.3],
        2024: [460.1, 475.3],
        2025: [434.1, 445.2],
    }
    grid_region_factors = PhAdorbGridRegion.model_validate(
        {
            "region_code": "DE",
            "region_name": "Germany",
            "description": "Germany",
            "hourly_CO2_factors": {
                "2023": [460.1, 469.3],
                "2024": [460.1, 475.3],
                "2025": [434.1, 445.2],
            },
        }
    )
    assert grid_region_factors.region_code == "DE"
    assert grid_region_factors.hourly_CO2_factors == hourly_CO2_factors
    assert grid_region_factors.get_CO2_factors_as_df().shape == (2, 3)
    assert grid_region_factors.get_CO2_factors_as_df().columns.tolist() == [
        2023,
        2024,
        2025,
    ]


def test_GridRegion_json_file():
    hourly_CO2_factors = {
        2023: [460.1, 469.3],
        2024: [460.1, 475.3],
        2025: [434.1, 445.2],
    }
    grid_region_factors = PhAdorbGridRegion(
        region_code="DE",
        region_name="Germany",
        description="Germany",
        hourly_CO2_factors=hourly_CO2_factors,
    )

    file_path = Path("test_grid_region.json")
    write_CO2_factors_to_json_file(file_path, grid_region_factors)

    # -- Read the JSON file data back in
    grid_region_factors_loaded = load_CO2_factors_from_json_file(
        Path("test_grid_region.json")
    )

    assert grid_region_factors == grid_region_factors_loaded
    assert grid_region_factors_loaded.get_CO2_factors_as_df().shape == (2, 3)
    assert grid_region_factors_loaded.get_CO2_factors_as_df().columns.tolist() == [
        2023,
        2024,
        2025,
    ]

    # -- Clean-up
    file_path.unlink()



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Calculate the annual costs for the ADORB analysis.

A.D.O.R.B. cost: Annualized De-carbonization Of Retrofitted Buildings cost - a “full-cost-accounted” 
annualized life-cycle cost metric for building projects. It includes the (annualized) direct costs of 
retrofit and maintenance, direct energy costs, a carbon cost for both operating and embodied/upfront 
greenhouse gas emissions, and a renewable-energy system-transition cost based on the required 
electrical service capacity.
"""

import logging
import pandas as pd

from ph_adorb.yearly_values import YearlyCost, YearlyPresentValueFactor

logger = logging.getLogger(__name__)

# -- Constants
# TODO: Support non-USA countries.
# TODO: Move these to be variables someplace...
USA_NUM_YEARS_TO_TRANSITION = 30
USA_NATIONAL_TRANSITION_COST = 4.5e12
NAMEPLATE_CAPACITY_INCREASE_GW = 1_600
USA_TRANSITION_COST_FACTOR = USA_NATIONAL_TRANSITION_COST / (NAMEPLATE_CAPACITY_INCREASE_GW * 1e9)


# ---------------------------------------------------------------------------------------


def present_value_factor(_year: int, _discount_rate: float) -> YearlyPresentValueFactor:
    """Calculate the present value factor for a given year."""
    rate = (1 + _discount_rate) ** (_year + 1)
    return YearlyPresentValueFactor(rate, _year + 1)


def energy_purchase_cost_PV(
    _pv_factor: YearlyPresentValueFactor, _annual_cost_electric: float, _annual_cost_gas: float
) -> float:
    """Calculate the total direct energy cost for a given year."""
    logger.info(f"energy_purchase_cost_PV(year={_pv_factor.year}, factor={_pv_factor.factor :.3f})")

    if _pv_factor.factor == 0:
        return 0.0

    annual_energy_cost = _annual_cost_electric + _annual_cost_gas
    annual_energy_cost_PV = annual_energy_cost / _pv_factor.factor

    logger.debug(
        f"Energy Actual Cost: ${_annual_cost_electric :,.0f}[Elec.] + ${_annual_cost_gas :,.0f}[Gas] = ${annual_energy_cost :,.0f}"
    )
    logger.debug(
        f"Energy PV Cost: ${annual_energy_cost :,.0f} / {_pv_factor.factor :.3f} = PV${annual_energy_cost_PV :,.0f}"
    )

    return annual_energy_cost_PV


def energy_CO2_cost_PV(
    _pv_factor: YearlyPresentValueFactor,
    _future_annual_CO2_electric: list[float],
    _annual_CO2_gas: float,
    _price_of_carbon: float,
) -> float:
    """Calculate the total operational carbon cost for a given year."""
    logger.info(f"energy_CO2_cost_PV(year={_pv_factor.year}, factor={_pv_factor.factor :.3f})")

    if _pv_factor.factor == 0:
        return 0.0

    annual_elec_CO2 = _future_annual_CO2_electric[_pv_factor.year - 1]
    annual_CO2 = annual_elec_CO2 + _annual_CO2_gas
    annual_CO2_cost = annual_CO2 * _price_of_carbon
    annual_CO2_cost_PV = annual_CO2_cost / _pv_factor.factor

    logger.debug(
        f"Energy CO2 Emissions: {annual_elec_CO2 :,.0f}[Elec.] + {_annual_CO2_gas :,.0f}[Gas] = {annual_CO2 :,.0f}"
    )
    logger.debug(
        f"Energy CO2 Actual Cost: {annual_CO2 :,.0f} * ${_price_of_carbon :,.2f}/unit = ${annual_CO2_cost :,.0f}"
    )
    logger.debug(
        f"Energy CO2 PV Cost [{_pv_factor.year}]: ${annual_CO2_cost :,.0f} / {_pv_factor.factor :.3f} = PV${annual_CO2_cost_PV :,.0f}"
    )

    return annual_CO2_cost_PV


def measure_purchase_cost_PV(
    _pv_factor: YearlyPresentValueFactor, _carbon_measure_yearly_purchase_costs: list[YearlyCost]
) -> float:
    """Calculate the total Measure purchase, install and maintenance cost for a single year."""
    logger.info(f"measure_purchase_cost_PV(year={_pv_factor.year}, factor={_pv_factor.factor :.3f})")

    if _pv_factor == 0:
        return 0.0

    measure_costs = [
        measure.cost for measure in _carbon_measure_yearly_purchase_costs if measure.year == _pv_factor.year - 1
    ]
    if not measure_costs:
        return 0.0

    total_measure_cost = sum(measure_costs)
    total_measure_PV_cost = total_measure_cost / _pv_factor.factor

    logging.debug(f"Measure Actual Costs: {[f'${_ :,.0f}' for _ in measure_costs]} = {total_measure_cost :,.0f}")
    logging.debug(
        f"Measure PV Cost: ${total_measure_cost :,.0f} / {_pv_factor.factor :.3f} = PV${total_measure_PV_cost :,.0f}"
    )

    return total_measure_PV_cost


def measure_CO2_cost_PV(
    _pv_factor: YearlyPresentValueFactor, _carbon_measure_embodied_CO2_yearly_costs: list[YearlyCost]
) -> float:
    """Calculate the total Measure embodied CO2 cost for a given year."""
    logger.info(f"measure_CO2_cost_PV(year={_pv_factor.year}, factor={_pv_factor.factor :.3f})")

    # TODO: What is this factor for? Why do we multiply by it?
    FACTOR = 0.75

    if _pv_factor.factor == 0:
        return 0.0

    measure_costs = [
        yearly_cost.cost
        for yearly_cost in _carbon_measure_embodied_CO2_yearly_costs
        if yearly_cost.year == _pv_factor.year - 1
    ]

    if not measure_costs:
        return 0.0

    total_measure_cost = sum(measure_costs)
    total_measure_PV_cost = sum(FACTOR * (cost / _pv_factor.factor) for cost in measure_costs)

    logging.debug(f"Measure CO2 Actual Cost: {[f'${_ :,.0f}' for _ in measure_costs]} = {total_measure_cost :,.0f}")
    logging.debug(
        f"Measure CO2 PV Cost: ${total_measure_cost :,.0f} / {_pv_factor.factor :.3f} = PV${total_measure_PV_cost :,.0f}"
    )

    return total_measure_PV_cost


def grid_transition_cost_PV(_pv_factor: YearlyPresentValueFactor, _peak_electrical_W: float) -> float:
    """Calculate the total grid transition cost for a given year."""
    logger.info(f"grid_transition_PV_cost(year={_pv_factor.year}, factor={_pv_factor.factor :.3f})")

    if _pv_factor.year > USA_NUM_YEARS_TO_TRANSITION:
        year_transition_cost_factor = 0  # $/Watt-yr
    else:
        # TODO: Support non-USA countries.
        year_transition_cost_factor = USA_TRANSITION_COST_FACTOR / USA_NUM_YEARS_TO_TRANSITION  # linear transition <- ?

    if _pv_factor.factor == 0:
        return 0.0

    transition_cost = year_transition_cost_factor * _peak_electrical_W
    transition_PV_cost = transition_cost / _pv_factor.factor

    logger.debug(
        f"Transition Cost [{_pv_factor.year}]: {year_transition_cost_factor: .5f} * {_peak_electrical_W :,.0f} W = ${transition_cost :,.0f}"
    )
    logger.debug(
        f"Transition PV Cost [{_pv_factor.year}]: ${transition_cost :,.0f} / {_pv_factor.factor :.3f} = PV${transition_PV_cost :,.0f}"
    )

    return transition_PV_cost


def calculate_annual_ADORB_costs(
    _analysis_duration_years: int,
    _annual_total_cost_electric: float,
    _annual_total_cost_gas: float,
    _annual_hourly_CO2_electric: list[float],
    _annual_total_CO2_gas: float,
    _all_yearly_install_costs: list[YearlyCost],
    _all_yearly_embodied_kgCO2: list[YearlyCost],
    _peak_electrical_W: float,
    _price_of_carbon: float,
) -> pd.DataFrame:
    """Returns a DataFrame with the yearly costs from the ADORB analysis."""

    # --  Define the column names
    columns = [
        "pv_direct_energy",
        "pv_operational_CO2",
        "pv_direct_MR",
        "pv_embodied_CO2",
        "pv_e_trans",
    ]

    # -- Create the row data
    rows: list[pd.Series] = []
    for n in range(0, _analysis_duration_years):
        logger.info(f"Calculating year {n} costs:")

        new_row: pd.Series[float] = pd.Series(
            {
                columns[0]: energy_purchase_cost_PV(
                    present_value_factor(n, 0.02), _annual_total_cost_electric, _annual_total_cost_gas
                ),
                columns[1]: energy_CO2_cost_PV(
                    present_value_factor(n, 0.075), _annual_hourly_CO2_electric, _annual_total_CO2_gas, _price_of_carbon
                ),
                columns[2]: measure_purchase_cost_PV(present_value_factor(n, 0.02), _all_yearly_install_costs),
                columns[3]: measure_CO2_cost_PV(present_value_factor(n, 0.0), _all_yearly_embodied_kgCO2),
                columns[4]: grid_transition_cost_PV(present_value_factor(n, 0.02), _peak_electrical_W),
            }
        )
        rows.append(new_row)

    return pd.DataFrame(rows, columns=columns)



import pandas as pd

from ph_adorb.grid_region import PhAdorbGridRegion
from ph_adorb.variant import calc_annual_total_electric_cost, calc_annual_hourly_electric_CO2


def test_get_annual_electric_cost():
    result = calc_annual_total_electric_cost(
        _purchased_electricity_kwh=100.0,
        _sold_electricity_kwh=0.0,
        _electric_purchase_price_per_kwh=0.1,
        _electric_sell_price_per_kwh=0.05,
        _electric_annual_base_price=100,
    )
    assert result == 110.0


def test_get_hourly_electric_CO2():

    grid_region_factors = PhAdorbGridRegion(
        region_code="Test",
        region_name="Germany",
        description="Germany",
        hourly_CO2_factors={k: v for k, v in zip(range(2023, 2023 + 89), [[0.0] * 100] * 89)},
    )
    result = calc_annual_hourly_electric_CO2(
        _hourly_purchased_electricity_kwh=[100.0, 100.0, 0.0], _grid_region=grid_region_factors
    )
    assert result == [0.0] * 89


# TODO: Add tests for the remaining functions.



from pathlib import Path

from ph_adorb.national_emissions import (
    PhAdorbNationalEmissions,
    load_national_emissions_from_json_file,
    write_national_emissions_to_json_file,
)


def test_NationalEmissions():
    national_emissions = PhAdorbNationalEmissions(
        country_name="Germany",
        us_trading_rank=4,
        GDP_million_USD=4.2,
        CO2_MT=2.3,
        kg_CO2_per_USD=0.5,
    )
    assert national_emissions.country_name == "Germany"
    assert national_emissions.us_trading_rank == 4
    assert national_emissions.GDP_million_USD == 4.2
    assert national_emissions.CO2_MT == 2.3
    assert national_emissions.kg_CO2_per_USD == 0.5
    assert national_emissions.model_dump() == {
        "country_name": "Germany",
        "us_trading_rank": 4,
        "GDP_million_USD": 4.2,
        "CO2_MT": 2.3,
        "kg_CO2_per_USD": 0.5,
    }


def test_NationalEmissions_to_json():
    national_emissions = PhAdorbNationalEmissions(
        country_name="Germany",
        us_trading_rank=4,
        GDP_million_USD=4.2,
        CO2_MT=2.3,
        kg_CO2_per_USD=0.5,
    )
    json_str = national_emissions.model_dump_json()
    assert (
        json_str
        == '{"country_name":"Germany","us_trading_rank":4,"GDP_million_USD":4.2,"CO2_MT":2.3,"kg_CO2_per_USD":0.5}'
    )


def test_NationalEmissions_from_json():
    national_emissions = PhAdorbNationalEmissions.model_validate(
        {
            "country_name": "Germany",
            "us_trading_rank": 4,
            "GDP_million_USD": 4.2,
            "CO2_MT": 2.3,
            "kg_CO2_per_USD": 0.5,
        }
    )
    assert national_emissions.country_name == "Germany"
    assert national_emissions.us_trading_rank == 4
    assert national_emissions.GDP_million_USD == 4.2
    assert national_emissions.CO2_MT == 2.3
    assert national_emissions.kg_CO2_per_USD == 0.5
    assert national_emissions.model_dump() == {
        "country_name": "Germany",
        "us_trading_rank": 4,
        "GDP_million_USD": 4.2,
        "CO2_MT": 2.3,
        "kg_CO2_per_USD": 0.5,
    }


def test_national_emissions_json_file():
    national_emissions = PhAdorbNationalEmissions(
        country_name="Germany",
        us_trading_rank=4,
        GDP_million_USD=4.2,
        CO2_MT=2.3,
        kg_CO2_per_USD=0.5,
    )
    file_path = Path("test.json")
    write_national_emissions_to_json_file(file_path, {"Germany": national_emissions})

    # -- Read the JSON Emissions back in
    emissions = load_national_emissions_from_json_file(file_path)
    assert len(emissions) == 1
    assert emissions["Germany"] == national_emissions

    # -- Cleanup
    file_path.unlink()



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Create a new Phius ADORB Variant from a Honeybee-Model."""

from collections import defaultdict
from pathlib import Path
from typing import Union

from honeybee.model import Model
from honeybee_energy.construction.opaque import OpaqueConstruction
from honeybee_energy.construction.window import WindowConstruction
from honeybee_energy.generator.pv import PVProperties
from honeybee_energy.load.lighting import Lighting
from honeybee_energy.load.process import Process
from honeybee_energy.properties.extension import (
    AllAirSystemProperties,
    DOASSystemProperties,
    HeatCoolSystemProperties,
    IdealAirSystemProperties,
)
from honeybee_energy.properties.model import ModelEnergyProperties
from honeybee_energy.properties.room import RoomEnergyProperties
from honeybee_energy.properties.shade import ShadeEnergyProperties

AnyHvacSystemProperties = Union[
    AllAirSystemProperties, DOASSystemProperties, HeatCoolSystemProperties, IdealAirSystemProperties
]
from honeybee_energy_revive.hvac.equipment import PhiusReviveHVACEquipment
from honeybee_energy_revive.properties.construction.opaque import OpaqueConstructionReviveProperties
from honeybee_energy_revive.properties.generator.pv import PVPropertiesReviveProperties
from honeybee_energy_revive.properties.hvac.allair import AllAirSystemReviveProperties
from honeybee_energy_revive.properties.hvac.doas import DOASSystemReviveProperties
from honeybee_energy_revive.properties.hvac.heatcool import HeatCoolSystemReviveProperties
from honeybee_energy_revive.properties.hvac.idealair import IdealAirSystemReviveProperties
from honeybee_energy_revive.properties.load.lighting import LightingReviveProperties
from honeybee_energy_revive.properties.load.process import ProcessReviveProperties
from honeybee_revive.properties.model import ModelReviveProperties

AnyHvacSystemReviveProperties = Union[
    AllAirSystemReviveProperties,
    DOASSystemReviveProperties,
    HeatCoolSystemReviveProperties,
    IdealAirSystemReviveProperties,
]

from ph_adorb.constructions import PhAdorbConstruction, PhAdorbConstructionCollection
from ph_adorb.ep_sql_file import DataFileSQL
from ph_adorb.equipment import PhAdorbEquipment, PhAdorbEquipmentCollection, PhAdorbEquipmentType
from ph_adorb.fuel import PhAdorbFuel, PhAdorbFuelType
from ph_adorb.grid_region import PhAdorbGridRegion, load_CO2_factors_from_json_file
from ph_adorb.measures import PhAdorbCO2MeasureCollection, PhAdorbCO2ReductionMeasure, CO2MeasureType
from ph_adorb.national_emissions import PhAdorbNationalEmissions
from ph_adorb.variant import PhAdorbVariant

# TODO: Add error / warning messages if GridRegion and NationalEmissions are not set in the HB-Model.


def get_hb_model_construction_quantities(_hb_model: Model) -> dict[str, float]:
    """Return a dictionary of total construction quantities (areas) from the HB-Model."""
    construction_quantities_ = defaultdict(float)
    for face in _hb_model.faces:
        for ap in face.apertures:
            construction_quantities_[ap.properties.energy.construction.identifier] += ap.area
        construction_quantities_[face.properties.energy.construction.identifier] += face.area
    return construction_quantities_


def get_PhAdorbGridRegion_from_hb_model(_hb_model_prop: ModelReviveProperties) -> PhAdorbGridRegion:
    """Get the Grid Region name from the HB-Model and load the data from file."""
    grid_region_data_path = Path(_hb_model_prop.grid_region.filepath)
    return load_CO2_factors_from_json_file(grid_region_data_path)


def get_PhAdorbNationalEmissions_from_hb_mode(_hb_model_prop: ModelReviveProperties) -> PhAdorbNationalEmissions:
    """Get the National Emissions data from the HB-Model."""
    return PhAdorbNationalEmissions(**_hb_model_prop.national_emissions_factors.to_dict())


def get_PhAdorbCO2Measures_from_hb_model(_hb_model_prop: ModelReviveProperties) -> PhAdorbCO2MeasureCollection:
    """Get all of the CO2 Reduction Measures from the HB-Model."""
    measure_collection_ = PhAdorbCO2MeasureCollection()
    for co2_measure in _hb_model_prop.co2_measures:
        measure_collection_.add_measure(
            PhAdorbCO2ReductionMeasure(
                measure_type=CO2MeasureType(co2_measure.measure_type),
                name=co2_measure.name,
                year=co2_measure.year,
                cost=co2_measure.cost,
                kg_CO2=co2_measure.kg_CO2,
                country_name=co2_measure.country_name,
                labor_fraction=co2_measure.labor_fraction,
            )
        )
    return measure_collection_


def convert_hb_construction(_hb_construction: OpaqueConstruction | WindowConstruction) -> PhAdorbConstruction:
    """Convert a Honeybee Opaque Construction to a Phius ADORB Construction."""
    hb_const_prop: OpaqueConstructionReviveProperties = getattr(_hb_construction.properties, "revive")
    return PhAdorbConstruction(
        display_name=_hb_construction.display_name,
        identifier=_hb_construction.identifier,
        CO2_kg_per_m2=hb_const_prop.kg_CO2_per_m2.value,
        cost_per_m2=hb_const_prop.cost_per_m2.value,
        lifetime_years=hb_const_prop.lifetime_years,
        labor_fraction=hb_const_prop.labor_fraction,
    )


def get_PhAdorbConstructions_from_hb_model(_hb_model: Model) -> PhAdorbConstructionCollection:
    """Return a ConstructionCollection with all of the Constructions from the HB-Model."""
    construction_areas = get_hb_model_construction_quantities(_hb_model)

    construction_collection = PhAdorbConstructionCollection()
    model_prop: ModelEnergyProperties = getattr(_hb_model.properties, "energy")
    for construction in model_prop.constructions:
        new_construction = convert_hb_construction(construction)
        new_construction.area_m2 = construction_areas[construction.identifier]
        construction_collection.add_construction(new_construction)

    return construction_collection


def convert_hb_process_load(process_load: Process) -> PhAdorbEquipment:
    """Convert a Honeybee Process-Load to a Phius ADORB Appliance Equipment."""
    process_prop: ProcessReviveProperties = getattr(process_load.properties, "revive")
    return PhAdorbEquipment(
        name=process_load.display_name,
        equipment_type=PhAdorbEquipmentType.APPLIANCE,
        cost=process_prop.cost,
        lifetime_years=process_prop.lifetime_years,
        labor_fraction=process_prop.labor_fraction,
    )


def convert_hbe_lighting(_hb_lighting: Lighting) -> PhAdorbEquipment:
    """Convert a Honeybee-Energy Lighting to a Phius ADORB Lighting Equipment."""
    lighting_prop: LightingReviveProperties = getattr(_hb_lighting.properties, "revive")
    return PhAdorbEquipment(
        name=_hb_lighting.display_name,
        equipment_type=PhAdorbEquipmentType.LIGHTS,
        cost=lighting_prop.cost,
        lifetime_years=lighting_prop.lifetime_years,
        labor_fraction=lighting_prop.labor_fraction,
    )


def convert_hb_shade_pv(_hb_pv: PVProperties) -> PhAdorbEquipment:
    """Convert a Honeybee-Energy PVProperties to a Phius ADORB PV Equipment."""
    pv_prop_revive: PVPropertiesReviveProperties = getattr(_hb_pv.properties, "revive")
    return PhAdorbEquipment(
        name=_hb_pv.display_name,
        equipment_type=PhAdorbEquipmentType.PV_ARRAY,
        cost=pv_prop_revive.cost,
        lifetime_years=pv_prop_revive.lifetime_years,
        labor_fraction=pv_prop_revive.labor_fraction,
    )


def convert_hb_hvac_equipment(_hb_hvac_equip: PhiusReviveHVACEquipment) -> PhAdorbEquipment:
    """Convert a Honeybee-Energy HVAC Equipment to a Phius ADORB HVAC Equipment."""
    return PhAdorbEquipment(
        name=_hb_hvac_equip.display_name,
        equipment_type=PhAdorbEquipmentType.MECHANICAL,
        cost=_hb_hvac_equip.cost,
        lifetime_years=_hb_hvac_equip.lifetime_years,
        labor_fraction=_hb_hvac_equip.labor_fraction,
    )


def get_PhAdorbEquipment_from_hb_model(_hb_model: Model) -> PhAdorbEquipmentCollection:
    """Return a EquipmentCollection with all of the Equipment (Appliances, HVAC, etc...) from the HB-Model."""

    equipment_collection_ = PhAdorbEquipmentCollection()

    for room in _hb_model.rooms:
        room_prop: RoomEnergyProperties = getattr(room.properties, "energy")

        # -- Add all of the Appliances from all of the HB-Rooms
        for process_load in room_prop.process_loads:
            equipment_collection_.add_equipment(convert_hb_process_load(process_load))

        # -- Add the room's lighting
        equipment_collection_.add_equipment(convert_hbe_lighting(room_prop.lighting))

        # -- Add the room's HVAC Equipment
        if not room_prop.hvac:
            continue
        hvac_props: AnyHvacSystemProperties = getattr(room_prop.hvac, "properties")
        hvac_prop_revive: AnyHvacSystemReviveProperties = getattr(hvac_props, "revive")
        for hb_hvac_equip in hvac_prop_revive.equipment_collection:
            equipment_collection_.add_equipment(convert_hb_hvac_equipment(hb_hvac_equip))

    # -- Add all the Model's Shades which have PV on them
    for shade in _hb_model.shades:
        shade_prop_e: ShadeEnergyProperties = getattr(shade.properties, "energy")
        if not shade_prop_e.pv_properties:
            continue
        equipment_collection_.add_equipment(convert_hb_shade_pv(shade_prop_e.pv_properties))

    return equipment_collection_


def get_PhAdorbFuels_from_hb_model(_hb_model: Model) -> tuple[PhAdorbFuel, PhAdorbFuel]:
    """Get the Electric and Natural-Gas Fuels from the HB-Model."""

    model_props: ModelReviveProperties = getattr(_hb_model.properties, "revive")
    hbrv_elec = model_props.fuels.get_fuel("ELECTRICITY")
    hb_nat_gas = model_props.fuels.get_fuel("NATURAL_GAS")

    electricity = PhAdorbFuel(
        fuel_type=PhAdorbFuelType.ELECTRICITY,
        purchase_price_per_kwh=hbrv_elec.purchase_price_per_kwh,
        sale_price_per_kwh=hbrv_elec.sale_price_per_kwh,
        annual_base_price=hbrv_elec.annual_base_price,
        used=True,
    )
    gas = PhAdorbFuel(
        fuel_type=PhAdorbFuelType.NATURAL_GAS,
        purchase_price_per_kwh=hb_nat_gas.purchase_price_per_kwh,
        sale_price_per_kwh=hb_nat_gas.sale_price_per_kwh,
        annual_base_price=hb_nat_gas.annual_base_price,
        used=True,
    )

    return electricity, gas


def get_PhAdorbVariant_from_hb_model(_hb_model: Model, _results_sql_file_path: Path) -> PhAdorbVariant:
    """Convert the HB-Model to a new ReviveVariant object.

    Arguments:
    ----------
        * hb_model (HB_Model): The Honeybee Model to convert.

    Returns:
    --------
        * ReviveVariant: The ReviveVariant object.
    """

    # -----------------------------------------------------------------------------------
    # -- Load in the EnergyPlus Simulation Result .SQL data file
    ep_results_sql = DataFileSQL(source_file_path=_results_sql_file_path)

    # -----------------------------------------------------------------------------------
    # -- Create the actual Variant
    hb_model_properties: ModelReviveProperties = getattr(_hb_model.properties, "revive")
    electricity, gas = get_PhAdorbFuels_from_hb_model(_hb_model)

    try:
        revive_variant = PhAdorbVariant(
            name=_hb_model.display_name or "unnamed",
            total_purchased_gas_kwh=ep_results_sql.get_total_purchased_gas_kwh(),
            hourly_purchased_electricity_kwh=ep_results_sql.get_hourly_purchased_electricity_kwh(),
            total_sold_electricity_kwh=ep_results_sql.get_total_sold_electricity_kwh(),
            peak_electric_usage_W=ep_results_sql.get_peak_electric_watts(),
            electricity=electricity,
            gas=gas,
            grid_region=get_PhAdorbGridRegion_from_hb_model(hb_model_properties),
            national_emissions=get_PhAdorbNationalEmissions_from_hb_mode(hb_model_properties),
            analysis_duration=hb_model_properties.analysis_duration,
            envelope_labor_cost_fraction=hb_model_properties.envelope_labor_cost_fraction,
            measure_collection=get_PhAdorbCO2Measures_from_hb_model(hb_model_properties),
            construction_collection=get_PhAdorbConstructions_from_hb_model(_hb_model),
            equipment_collection=get_PhAdorbEquipment_from_hb_model(_hb_model),
        )
    except Exception as e:
        msg = (
            "An error occurred while reading data from the EnergyPlus SQL file? Please be sure that "
            "you have set all of the required output-variables before running the EnergyPlus simulation"
        )
        raise Exception(msg, e)

    return revive_variant



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""A script to calculate ADORB Costs from a Honeybee-Model-HBJSON-File, and outputs to CSV files.

This script is called from the command line with the following arguments:
    * [1] (str): The path to the HBJSON file to read in.
    * [2] (str): The path to the EnergyPlus SQL file to read in.
    * [3] (str): The path to the output Yearly CSV file.
    * [4] (str): The path to the output Cumulative CSV file.
    * [5] (str): The path to the output folder for the preview tables.
"""

import os
import shutil
import sys
from pathlib import Path
from collections import namedtuple
import logging
from logging import getLogger

from ph_adorb.from_HBJSON import create_variant, read_HBJSON_file
from ph_adorb.variant import calc_variant_yearly_ADORB_costs, calc_variant_cumulative_ADORB_costs


# Function to setup logger to output to a log file
def setup_logger(log_file_path: Path, _level: int = logging.INFO) -> logging.Logger:
    logging.basicConfig(
        filename=log_file_path, filemode="w", level=_level, format="%(asctime)s - %(levelname)s - %(message)s"
    )
    return getLogger(__name__)


class InputFileError(Exception):
    def __init__(self, path) -> None:
        self.msg = f"\nCannot find the specified HBJSON file:'{path}'"
        super().__init__(self.msg)


Filepaths = namedtuple("Filepaths", ["hbjson", "sql", "annual_csv", "cumulative_csv", "tables"])


def _remove_folder_and_contents(_folder: Path) -> None:
    """Remove all files in the specified folder."""
    print(f"\t>> Removing: {_folder}")
    shutil.rmtree(_folder)


def resolve_paths(_args: list[str]) -> Filepaths:
    """Sort out the file input and output paths. Make the output directory if needed.

    Arguments:
    ----------
        * _args (list[str]): sys.args list of input arguments.

    Returns:
    --------
        * Filepaths
    """

    assert len(_args) == 6, "Error: Incorrect number of arguments."

    # -----------------------------------------------------------------------------------
    # -- The HBJSON input file.
    hbjson_source_filepath = Path(_args[1])
    if not hbjson_source_filepath.exists():
        raise InputFileError(hbjson_source_filepath)

    # -----------------------------------------------------------------------------------
    # -- The EnergyPlus SQL input file.
    results_sql_file = Path(_args[2])
    if not results_sql_file.exists():
        raise InputFileError(results_sql_file)

    # -----------------------------------------------------------------------------------
    # -- Annual CSV output file:
    # -- If the folder of the target_csv_filepath does not exist, make it.
    target_annual_csv_filepath = Path(_args[3])
    target_dir = target_annual_csv_filepath.parent
    if not target_dir.exists():
        print(f"\t>> Creating the directory: {target_dir}")
        os.mkdir(target_dir)

    # -- If the target CSV already exists, delete it.
    if target_annual_csv_filepath.exists():
        print(f"\t>> Removing the existing CSV file: {target_annual_csv_filepath}")
        os.remove(target_annual_csv_filepath)

    # -----------------------------------------------------------------------------------
    # -- Cumulative CSV output file:
    target_cumulative_csv_filepath = Path(_args[4])
    target_dir = target_cumulative_csv_filepath.parent
    if not target_dir.exists():
        print(f"\t>> Creating the directory: {target_dir}")
        os.mkdir(target_dir)

    # -- If the target CSV already exists, delete it.
    if target_cumulative_csv_filepath.exists():
        print(f"\t>> Removing the existing CSV file: {target_cumulative_csv_filepath}")
        os.remove(target_cumulative_csv_filepath)

    # -----------------------------------------------------------------------------------
    # -- Preview-Tables output folder:
    target_tables_dir = Path(_args[5])
    if target_tables_dir.exists():
        _remove_folder_and_contents(target_tables_dir)

    print(f"\t>> Creating the directory: {target_tables_dir}")
    os.mkdir(target_tables_dir)

    return Filepaths(
        hbjson_source_filepath,
        results_sql_file,
        target_annual_csv_filepath,
        target_cumulative_csv_filepath,
        target_tables_dir,
    )


if __name__ == "__main__":
    print("- " * 50)
    print(f"\t>> Using Python: {sys.version}")
    print(f"\t>> Running the script: '{__file__.split('/')[-1]}'")
    print(f"\t>> With the arguments:")
    print("\n".join([f"\t\t{i} | {a}" for i, a in enumerate(sys.argv)]))

    # --- Input / Output file Path
    # -------------------------------------------------------------------------
    print("\t>> Resolving file paths...")
    file_paths = resolve_paths(sys.argv)

    # Setup logger to output to the same directory as the 'file_paths.annual_csv' location
    log_file_path = file_paths.annual_csv.parent / "calc_HBJSON_ADORB_costs.log"
    logger = setup_logger(log_file_path, logging.DEBUG)

    logger.info("Script started")

    print(f"\t>> Source HBJSON File: '{file_paths.hbjson}'")
    print(f"\t>> Source SQL File: '{file_paths.sql}'")
    print(f"\t>> Target CSV File (yearly): '{file_paths.annual_csv}'")
    print(f"\t>> Target CSV File (cumulative): '{file_paths.cumulative_csv}'")
    print(f"\t>> Target Tables Output Folder: '{file_paths.tables}'")

    # --- Read in the existing HB-JSON-File
    # -------------------------------------------------------------------------
    print(f"\t>> Loading the Honeybee-Model from the HBJSON file: {file_paths.hbjson}")
    hb_json_dict = read_HBJSON_file.read_hb_json_from_file(file_paths.hbjson)

    # -- Re-Build the Honeybee-Model from the HBJSON-Dict
    # -------------------------------------------------------------------------
    hb_model = read_HBJSON_file.convert_hbjson_dict_to_hb_model(hb_json_dict)
    print(f"\t>> Honeybee-Model '{hb_model.display_name}' successfully re-built from file.")

    # --- Generate the PH-ADORB-Variant from the Honeybee-Model
    revive_variant = create_variant.get_PhAdorbVariant_from_hb_model(hb_model, file_paths.sql)

    # --- Get the ADORB Costs for the PH-ADORB-Variant
    # -------------------------------------------------------------------------
    variant_yearly_ADORB_df = calc_variant_yearly_ADORB_costs(revive_variant, file_paths.tables)
    variant_cumulative_ADORB_df = calc_variant_cumulative_ADORB_costs(variant_yearly_ADORB_df)

    # --- Output the ADORB Costs to a CSV File
    # -------------------------------------------------------------------------
    variant_yearly_ADORB_df.to_csv(file_paths.annual_csv)
    variant_cumulative_ADORB_df.to_csv(file_paths.cumulative_csv)
    print("\t>> Done calculating the ADORB Costs. The CSV files have been saved.")
    print("- " * 50)
    logger.info("Script finished")
